package edu.sc.sasimulator;
import java.util.*;
import java.io.*;
/** 
 * @author Jose M. Vidal
 * @version $Revision: 1.5 $, $Date: 2000/04/14 00:58:09 $
 */
public class Simulator {

    /** The current simulation time */
    private long time;

    private Allocation currentAllocation;
    
    /**
       * Get the value of currentAllocation.
       * @return Value of currentAllocation.
       */
    public Allocation getCurrentAllocation() {return currentAllocation;}
    
    /**
       * Set the value of currentAllocation.
       * @param v  Value to assign to currentAllocation.
       */
    public void setCurrentAllocation(Allocation  v) {this.currentAllocation = v;}
    


    /** Standard Binary Heap implementation of a priority queue */
    class PriorityQueue {
	/** The array of events, it is a binary head. The top of the heap is at index 1. */
	Event[] events;

	/** Index of first available (free) index. */
	private int last;

	PriorityQueue(){
	    events = new Event[1000];
	    last = 1;
	}

	/** Add an event in proper place */
	void add(Event e){
	    int hole = last;
	    for ( ; (hole > 1) && e.isBefore(events[hole/2]) ; hole /= 2)
		events[hole] = events[hole/2];
	    events[hole] = e;
	    last++;

	}

	/** Return the even at the top */
	Event pop(){
	    if (last <= 1)
		return null;
	    int child;
	    int hole = 1;
	    Event top = events[1];
	    events[1] = events[--last];
	    Event tmp = events[1];
	    for (; hole* 2 < last ; hole = child){
		//		System.out.println("hole="+hole);
		child = hole * 2;
		if (child+1 < last && events[child+1].isBefore(events[child]))
		    child++;
		if (events[child].isBefore(tmp)){
		    events[hole] = events[child];
		    events[child] = null;
		}
		else 
		    break;
	    }
	    events[hole] = tmp;
	    return top;
	}
    }

    PriorityQueue pq;

    private Vector agents;
    /**
       * Get the value of agents.
       * @return Value of agents.
       */
    public Vector getAgents() {return agents;}
    
    /**
       * Set the value of agents.
       * @param v  Value to assign to agents.
       */
    public void setAgents(Vector  v) {this.agents = v;}
    
    private Vector services;
    /**
       * Get the value of services.
       * @return Value of services.
       */
    public Vector getServices() {return services;}
    
    /**
       * Set the value of services.
       * @param v  Value to assign to services.
       */
    public void setServices(Vector  v) {this.services = v;}
    

    private PrintStream fileout;
    /**
       * Get the value of fileout.
       * @return Value of fileout.
       */
    public PrintStream getFileout() {return fileout;}
    
    /**
       * Set the value of fileout.
       * @param v  Value to assign to fileout.
       */
    public void setFileout(PrintStream  v) {this.fileout = v;}
    
    private int steps;

    /** The alpha-values as a funtion of time, for this run */
    public double[] utilities;

    public Simulator(int s){
	pq = new PriorityQueue();
	currentAllocation = new Allocation();
	steps = s;
	utilities = new double[steps];
    }

    public void start(){
	pq = new PriorityQueue();
	currentAllocation = new Allocation();
	for (int i = 0 ; i < steps ; i++)
	    utilities[i] = 0;
    }
    
    /** Does an exhaustive search in order to find the best allocation. This will take 4<sup>|Services|</sup> time.
	* @param servs the vector of current services (those left to try), on the first call it should include all
	* @param a the current allocation. On the first call it should be a new (empty) allocation
	* @return the max utility */
    private Allocation getBestAllocation(Vector servs, Allocation a){
	Enumeration is = servs.elements();
	Service s = (Service)is.nextElement();
	servs.removeElement(s);
	double maxUtility = 0;
	Allocation bestAllocation = null;
	for (int sector=-1; sector < 3; ++sector){
	    a.removeServiceFromAll(s);
	    if (sector != -1) {
		for (Enumeration j = agents.elements(); j.hasMoreElements();){
		    Agent ag = (Agent)j.nextElement();
		    if (s.determineSector(ag) == sector){
			a.addServiceToAgent(s, ag);
		    }
		}
	    }
	    double u;
	    Allocation ca;
	    if (servs.size() == 0){
		ca = new Allocation(a);
	    }
	    else {
		Vector servicesCopy = (Vector)servs.clone();
		ca = getBestAllocation(servicesCopy, a);
	    }
	    u = getGlobalUtility(ca);
	    if (u > maxUtility){
		maxUtility = u;
		bestAllocation = ca;
	    }
	}
	return bestAllocation;
    }

    double maxUtility = -1;

    /** 
	@return the max utility, found by searching all possible allocations using getBestAllocation */
    public double getMaxUtility(){
	Allocation atemp = new Allocation(services);
	Allocation best = getBestAllocation((Vector)services.clone(), atemp);
	System.out.println("best=\n" + best.toString());
	maxUtility = getGlobalUtility(best);
	return maxUtility;
    }

    /**
       @return the utility of the current allocation */
    public double getCurrentUtility(){
	return getGlobalUtility(currentAllocation);
    }

	
    public void run(){
	int step = 0;
	double lastu = 0;
	for (Event e = pq.pop(); e != null && ++step < steps; e = pq.pop()){
	    time = e.time;
	    e.thing.doEvent(e.type);
	    Allocation a = currentAllocation;
	    double newu = getGlobalUtility();
// 	    utilities[step] = newu;
	    double u = newu - lastu;
	    double alpha = newu / maxUtility;
	    utilities[step] = alpha;
	    String s = step + "\t" + alpha;
	    //	    String s = lastu + "\t" + newu;
	    lastu = newu;
	    fileout.println(s);
	}
//  	System.out.println("last= " + getGlobalUtility(currentAllocation));
	if (!isAllocationOK(currentAllocation))
	    System.out.println("NOT OK");
// 	System.out.println(currentAllocation.toString());
    }

    private double getGlobalUtility(Allocation a){
	double u = 0;
	for (Enumeration i = agents.elements(); i.hasMoreElements();){
	    Agent ia = (Agent)i.nextElement();
	    u += ia.utility(a);
	}
	for (Enumeration i = services.elements(); i.hasMoreElements();){
	    Service is = (Service)i.nextElement();
	    u += is.utility(a);
	}

	return u;
    }

    private double getGlobalUtility(){
	return getGlobalUtility(currentAllocation);
    }

    /**
       @return true if a does not violate any constraints */
    private boolean isAllocationOK(Allocation a){
	for (Enumeration i = services.elements(); i.hasMoreElements();){
	    Service s = (Service)i.nextElement();
	    if (!(s.isOK(a)))
		return false;
	}
	return true;
    }
	
    
    public void addEvent(Thing t, long tm, int type){
	Event e = new Event(time + tm, type, t);
	pq.add(e);
    }

}
